@@ -3,8 +3,9 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2000-2015, University of Amsterdam
+    Copyright (c)  2000-2018, University of Amsterdam
                               VU University Amsterdam
+			      CWI, Amsterdam
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -35,9 +36,7 @@
 
 #define O_DEBUG 1
 
-#ifdef HAVE_CONFIG_H
 #include <config.h>
-#endif
 
 #ifdef __GNUC__
 # define alloca __builtin_alloca
@@ -76,6 +75,10 @@
 #define GET_H_ERRNO h_errno
 #endif
 
+#if defined(__linux__) || defined(__APPLE__)
+#define HAVE_IP_MREQN 1
+#endif
+
 static atom_t ATOM_reuseaddr;		/* "reuseaddr" */
 static atom_t ATOM_bindtodevice;	/* "bindtodevice" */
 static atom_t ATOM_broadcast;		/* "broadcast" */
@@ -88,7 +91,9 @@ static atom_t ATOM_atom;		/* "atom" */
 static atom_t ATOM_string;		/* "string" */
 static atom_t ATOM_codes;		/* "codes" */
 static atom_t ATOM_max_message_size;    /* "message_size" */
-
+static atom_t ATOM_file_no;		/* "file_no" */
+static atom_t ATOM_ip_add_membership;	/* "ip_add_membership" */
+static atom_t ATOM_ip_drop_membership;	/* "ip_drop_membership" */
 static functor_t FUNCTOR_socket1;	/* $socket(Id) */
 
 
@@ -178,6 +183,12 @@ pl_host_to_address(term_t Host, term_t Ip)
 }
 
 
+#ifndef HAVE_IP_MREQN
+#define ip_mreqn ip_mreq
+#define imr_address imr_interface
+#endif
+
+
 static foreign_t
 pl_setopt(term_t Socket, term_t opt)
 { int socket;
@@ -245,6 +256,40 @@ pl_setopt(term_t Socket, term_t opt)
     { if ( nbio_setopt(socket, TCP_NONBLOCK) == 0 )
 	return TRUE;
       return FALSE;
+#ifdef IP_ADD_MEMBERSHIP
+    } else if ( (a == ATOM_ip_add_membership || a == ATOM_ip_drop_membership)
+		&& arity >= 1 )
+    { struct ip_mreqn mreq;
+      term_t arg = PL_new_term_ref();
+      int opname = (a == ATOM_ip_add_membership) ? IP_ADD_MEMBERSHIP
+					         : IP_DROP_MEMBERSHIP;
+
+      _PL_get_arg(1, opt, arg);
+      memset(&mreq, 0, sizeof(mreq));
+      if ( !nbio_get_ip(arg, &mreq.imr_multiaddr) )
+	return PL_domain_error("ip", arg);
+      if ( arity >= 2 )
+      { _PL_get_arg(2, opt, arg);
+	if ( !nbio_get_ip(arg, &mreq.imr_address) )
+	  return PL_domain_error("ip", arg);
+      } else
+	mreq.imr_address.s_addr = htonl(INADDR_ANY);
+#ifdef HAVE_IP_MREQN
+      if ( arity == 3 )
+      { _PL_get_arg(3, opt, arg);
+	if ( !PL_get_integer(arg, &mreq.imr_ifindex) )
+	  return FALSE;
+      }
+#endif
+      if ( arity > 3 )
+	goto not_implemented;
+
+      if ( setsockopt(nbio_fd(socket), IPPROTO_IP, opname,
+		      (void*)&mreq, sizeof(mreq)) < 0 )
+	return nbio_error(GET_H_ERRNO, TCP_HERRNO);
+      else
+	return TRUE;
+#endif
     }
   }
 
@@ -253,6 +298,31 @@ not_implemented:
 }
 
 
+static foreign_t
+pl_getopt(term_t Socket, term_t opt)
+{ int socket;
+  atom_t a;
+  size_t arity;
+
+  if ( !tcp_get_socket(Socket, &socket) )
+    return FALSE;
+
+  if ( PL_get_name_arity(opt, &a, &arity) && arity >= 1 )
+  { term_t a1 = PL_new_term_ref();
+
+    _PL_get_arg(1, opt, a1);
+    if ( a == ATOM_file_no && arity == 1 )
+    { SOCKET s = nbio_fd(socket);
+
+      if ( s != -1 )
+	return PL_unify_integer(a1, s);
+      return FALSE;
+    }
+  }
+
+  return pl_error(NULL, 0, NULL, ERR_DOMAIN, opt, "socket_option");
+}
+
 #include "sockcommon.c"
 
 		 /*******************************
@@ -710,18 +780,21 @@ install_t
 install_socket(void)
 { nbio_init("socket");
 
-  ATOM_reuseaddr        = PL_new_atom("reuseaddr");
-  ATOM_bindtodevice     = PL_new_atom("bindtodevice");
-  ATOM_broadcast        = PL_new_atom("broadcast");
-  ATOM_nodelay	        = PL_new_atom("nodelay");
-  ATOM_dispatch	        = PL_new_atom("dispatch");
-  ATOM_nonblock	        = PL_new_atom("nonblock");
-  ATOM_infinite	        = PL_new_atom("infinite");
-  ATOM_as	        = PL_new_atom("as");
-  ATOM_atom	        = PL_new_atom("atom");
-  ATOM_string	        = PL_new_atom("string");
-  ATOM_codes	        = PL_new_atom("codes");
-  ATOM_max_message_size = PL_new_atom("max_message_size");
+  ATOM_reuseaddr          = PL_new_atom("reuseaddr");
+  ATOM_bindtodevice       = PL_new_atom("bindtodevice");
+  ATOM_broadcast          = PL_new_atom("broadcast");
+  ATOM_nodelay	          = PL_new_atom("nodelay");
+  ATOM_dispatch	          = PL_new_atom("dispatch");
+  ATOM_nonblock	          = PL_new_atom("nonblock");
+  ATOM_infinite	          = PL_new_atom("infinite");
+  ATOM_as	          = PL_new_atom("as");
+  ATOM_atom	          = PL_new_atom("atom");
+  ATOM_string	          = PL_new_atom("string");
+  ATOM_codes	          = PL_new_atom("codes");
+  ATOM_max_message_size   = PL_new_atom("max_message_size");
+  ATOM_file_no		  = PL_new_atom("file_no");
+  ATOM_ip_add_membership  = PL_new_atom("ip_add_membership");
+  ATOM_ip_drop_membership = PL_new_atom("ip_drop_membership");
 
   FUNCTOR_socket1 = PL_new_functor(PL_new_atom("$socket"), 1);
 
@@ -733,6 +806,7 @@ install_socket(void)
   PL_register_foreign("tcp_socket",           1, tcp_socket,          0);
   PL_register_foreign("tcp_close_socket",     1, pl_close_socket,     0);
   PL_register_foreign("tcp_setopt",           2, pl_setopt,           0);
+  PL_register_foreign("tcp_getopt",           2, pl_getopt,           0);
   PL_register_foreign("tcp_host_to_address",  2, pl_host_to_address,  0);
   PL_register_foreign("gethostname",          1, pl_gethostname,      0);
 #ifdef __WINDOWS__
