@@ -573,7 +573,11 @@ tokenizeA(const char *in, size_t len,
       }
 
       if ( !(*call)((const char*)st, s-st, type, closure) )
-	return FALSE;
+      { while(s<se && iswalnum(*s))
+	  s++;
+	if ( !(*call)((const char*)st, s-st, TOK_WORD, closure) )
+	  return FALSE;
+      }
     } else if ( iswalnum(*s) )
     { while(s<se && iswalnum(*s))
 	s++;
@@ -639,7 +643,12 @@ tokenizeW(const wchar_t *in, size_t len,
       }
 
       if ( !(*call)((const wchar_t*)st, s-st, type, closure) )
-	return FALSE;
+      { while(s<se && iswalnum(*s))
+	  s++;
+	if ( !(*call)((const wchar_t*)st, s-st, TOK_WORD, closure) )
+	  return FALSE;
+      }
+
     } else if ( iswalnum(*s) )
     { while(s<se && iswalnum(*s))
 	s++;
@@ -761,7 +770,8 @@ pl_tokenize(term_t text, term_t tokens)
   } else if ( PL_get_wchars(text, &len, &ws, CVT_ALL|CVT_EXCEPTION) )
   { if ( !tokenizeW(ws, len, unify_tokenW, &l) )
       return FALSE;
-  }
+  } else
+    return FALSE;
 
   return PL_unify_nil(l.tail);
 }
