@@ -1,33 +1,49 @@
-/*
-    pqConsole    : interfacing SWI-Prolog and Qt
+/*  Part of SWI-Prolog interface to Qt
 
-    Author       : Carlo Capelli
-    E-mail       : cc.carlo.cap@gmail.com
-    Copyright (C): 2013, Carlo Capelli
+    Author:        Carlo Capelli
+    E-mail:        cc.carlo.cap@gmail.com
+    Copyright (c)  2013-2015, Carlo Capelli
+    All rights reserved.
 
-    This library is free software; you can redistribute it and/or
-    modify it under the terms of the GNU Lesser General Public
-    License as published by the Free Software Foundation; either
-    version 2.1 of the License, or (at your option) any later version.
+    Redistribution and use in source and binary forms, with or without
+    modification, are permitted provided that the following conditions
+    are met:
 
-    This library is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-    Lesser General Public License for more details.
+    1. Redistributions of source code must retain the above copyright
+       notice, this list of conditions and the following disclaimer.
 
-    You should have received a copy of the GNU Lesser General Public
-    License along with this library; if not, write to the Free Software
-    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
+    2. Redistributions in binary form must reproduce the above copyright
+       notice, this list of conditions and the following disclaimer in
+       the documentation and/or other materials provided with the
+       distribution.
+
+    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
+    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
+    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
+    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
+    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
+    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
+    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
+    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+    POSSIBILITY OF SUCH DAMAGE.
 */
 
 #include <SWI-Stream.h>
+#include <signal.h>
+
 #include "Swipl_IO.h"
 #include "do_events.h"
 #include "PREDICATE.h"
 #include "Completion.h"
 #include "Preferences.h"
 #include "pqMainWindow.h"
-#include <signal.h>
+
+#include "blockSig.h"
+#include "ParenMatching.h"
+#include <QTextBlock>
 
 #include <QTime>
 #include <QRegExp>
@@ -149,6 +165,9 @@ void ConsoleEdit::setup() {
     connect(this, SIGNAL(cursorPositionChanged()), this, SLOT(onCursorPositionChanged()));
 
     connect(this, SIGNAL(sig_run_function(pfunc)), this, SLOT(run_function(pfunc)));
+
+    // so far,
+    connect(this, SIGNAL(selectionChanged()), this, SLOT(selectionChanged()));
 }
 
 /** strict control on keyboard events required
@@ -234,7 +253,7 @@ void ConsoleEdit::keyPressEvent(QKeyEvent *event) {
 
     case Key_Up:
         down = false;
-        // fall throu
+        // fall through
     case Key_Down:
         if (!ctrl) {
             // naive history handler
@@ -299,7 +318,7 @@ void ConsoleEdit::keyPressEvent(QKeyEvent *event) {
             PL_thread_raise(thids[0], SIGINT);
             return;
         }
-        // fall throu
+        // fall through
 
     default:
         accept = editable || event->matches(QKeySequence::Copy);
@@ -591,7 +610,7 @@ void ConsoleEdit::user_prompt(int threadId, bool tty) {
 
     is_tty = tty;
 
-    Completion::helpidx();
+    Completion::setup();
 
     QTextCursor c = textCursor();
     c.movePosition(QTextCursor::End);
@@ -876,11 +895,11 @@ void ConsoleEdit::query_run(QString module, QString call) {
 }
 
 ConsoleEdit::exec_sync::exec_sync(int timeout_ms) : timeout_ms(timeout_ms) {
-    stop_ = CT;
+    stop_ = QThread::currentThread();
     go_ = 0;
 }
 void ConsoleEdit::exec_sync::stop() {
-    Q_ASSERT(CT == stop_);
+    Q_ASSERT(QThread::currentThread() == stop_);
     for ( ; ; ) {
         {   QMutexLocker lk(&sync);
             if (go_)
@@ -893,7 +912,7 @@ void ConsoleEdit::exec_sync::stop() {
 void ConsoleEdit::exec_sync::go() {
     Q_ASSERT(go_ == 0);
     Q_ASSERT(stop_ != 0);
-    auto t = CT;
+    auto t = QThread::currentThread();
     if (stop_ != t) {
         QMutexLocker lk(&sync);
         go_ = t;
@@ -914,3 +933,43 @@ void ConsoleEdit::html_write(QString html) {
     c.movePosition(c.End);
     c.insertHtml(html);
 }
+
+void ConsoleEdit::selectionChanged()
+{
+    if (sel_check_timing.isValid()) {
+        if (sel_check_timing.elapsed() < 100)
+            return;
+        sel_check_timing.restart();
+    }
+    else
+        sel_check_timing.start();
+
+    blockSig bs(this);
+
+    foreach (ExtraSelection s, extraSelections())
+        s.cursor.setCharFormat(s.format);
+    extraSelections().clear();
+
+    QTextCursor c = textCursor();
+    if (c.hasSelection()) {
+        QString csel = c.selectedText();
+        QList<ExtraSelection> lsel;
+        QTextCharFormat bold = ParenMatching::range::bold();
+
+        QTextCursor cfirst = cursorForPosition(QPoint(0, 0));
+        if (!cfirst.isNull()) {
+            while (c.block().position() > cfirst.block().position())
+                c.movePosition(c.Up);
+            c.movePosition(c.Up);
+            for ( ; ; ) {
+                c = document()->find(csel, c, QTextDocument::FindCaseSensitively);
+                if (c.isNull() || !c.block().isVisible())
+                    break;
+                lsel.append(ExtraSelection {c, c.blockCharFormat()});
+                c.setCharFormat(bold);
+            }
+        }
+
+        setExtraSelections(lsel);
+    }
+}
