@@ -68,7 +68,7 @@ extern "C" {
 /* PLVERSION_TAG: a string, normally "", but for example "rc1" */
 
 #ifndef PLVERSION
-#define PLVERSION 70602
+#define PLVERSION 70725
 #endif
 #ifndef PLVERSION_TAG
 #define PLVERSION_TAG ""
@@ -394,8 +394,8 @@ PL_EXPORT(int)		PL_predicate_info(predicate_t pred,
 PL_EXPORT(qid_t)	PL_open_query(module_t m, int flags,
 				      predicate_t pred, term_t t0);
 PL_EXPORT(int)		PL_next_solution(qid_t qid) WUNUSED;
-PL_EXPORT(void)		PL_close_query(qid_t qid);
-PL_EXPORT(void)		PL_cut_query(qid_t qid);
+PL_EXPORT(int)		PL_close_query(qid_t qid);
+PL_EXPORT(int)		PL_cut_query(qid_t qid);
 PL_EXPORT(qid_t)	PL_current_query(void);
 
 			/* Simplified (but less flexible) call-back */
@@ -721,8 +721,16 @@ PL_EXPORT(char *)	PL_cwd(char *buf, size_t buflen);
 		 *    QUINTUS/SICSTUS WRAPPER	*
 		 *******************************/
 
+PL_EXPORT(int)		PL_cvt_i_char(term_t p, char *c);
+PL_EXPORT(int)		PL_cvt_i_uchar(term_t p, unsigned char *c);
+PL_EXPORT(int)		PL_cvt_i_short(term_t p, short *s);
+PL_EXPORT(int)		PL_cvt_i_ushort(term_t p, unsigned short *s);
 PL_EXPORT(int)		PL_cvt_i_int(term_t p, int *c);
+PL_EXPORT(int)		PL_cvt_i_uint(term_t p, unsigned int *c);
 PL_EXPORT(int)		PL_cvt_i_long(term_t p, long *c);
+PL_EXPORT(int)		PL_cvt_i_ulong(term_t p, unsigned long *c);
+PL_EXPORT(int)		PL_cvt_i_int64(term_t p, int64_t *c);
+PL_EXPORT(int)		PL_cvt_i_uint64(term_t p, uint64_t *c);
 PL_EXPORT(int)		PL_cvt_i_size_t(term_t p, size_t *c);
 PL_EXPORT(int)		PL_cvt_i_float(term_t p, double *c);
 PL_EXPORT(int)		PL_cvt_i_single(term_t p, float *c);
@@ -928,6 +936,16 @@ PL_EXPORT(int)		PL_cleanup(int status);
 PL_EXPORT(void)		PL_cleanup_fork();
 PL_EXPORT(int)		PL_halt(int status);
 
+		 /*******************************
+		 *	  DYNAMIC LINKING	*
+		 *******************************/
+
+PL_EXPORT(void *)	PL_dlopen(const char *file, int flags);
+PL_EXPORT(const char *) PL_dlerror(void);
+PL_EXPORT(void *)	PL_dlsym(void *handle, char *symbol);
+PL_EXPORT(int)		PL_dlclose(void *handle);
+
+
 		 /*******************************
 		 *      INPUT/PROMPT/ETC	*
 		 *******************************/
@@ -1080,15 +1098,20 @@ PL_EXPORT(intptr_t)	PL_query(int);	/* get information from Prolog */
 #define PL_THREAD_NO_DEBUG	0x01	/* Start thread in nodebug mode */
 #define PL_THREAD_NOT_DETACHED	0x02	/* Allow Prolog to join */
 
+typedef enum
+{ PL_THREAD_CANCEL_FAILED = FALSE,	/* failed to cancel; try abort */
+  PL_THREAD_CANCEL_JOINED = TRUE,	/* cancelled and joined */
+  PL_THREAD_CANCEL_MUST_JOIN		/* cancelled, must join */
+} rc_cancel;
+
 typedef struct
-{ long	    local_size;			/* Stack sizes (Kbytes) */
-  long	    global_size;
-  long	    trail_size;
-  long	    argument_size;
+{ size_t    stack_limit;		/* Total stack limit (bytes) */
+  size_t    table_space;		/* Total tabling space limit (bytes) */
   char *    alias;			/* alias name */
-  int	  (*cancel)(int id);		/* cancel function */
+  rc_cancel (*cancel)(int id);		/* cancel function */
   intptr_t  flags;			/* PL_THREAD_* flags */
-  void *    reserved[4];		/* reserved for extensions */
+  size_t    max_queue_size;		/* Max size of associated queue */
+  void *    reserved[3];		/* reserved for extensions */
 } PL_thread_attr_t;
 
 
@@ -1106,6 +1129,7 @@ PL_EXPORT(int)	PL_thread_raise(int tid, int sig);
 PL_EXPORT(int)	PL_w32thread_raise(DWORD dwTid, int sig);
 PL_EXPORT(int)	PL_wait_for_console_input(void *handle);
 PL_EXPORT(int)	PL_w32_wrap_ansi_console(void);
+PL_EXPORT(const char*) PL_w32_running_under_wine(void);
 #endif
 
 		 /*******************************
