@@ -209,7 +209,9 @@ print_backtrace(int last)		/* 1..SAVE_TRACES */
 
 void
 bstore_print_backtrace_named(btrace *bt, const char *why)
-{ if ( bt )
+{ int done = 0;
+
+  if ( bt )
   { int me = bt->current-1;
 
     for(;;)
@@ -217,14 +219,15 @@ bstore_print_backtrace_named(btrace *bt, const char *why)
 	me += SAVE_TRACES;
       if ( bt->dumps[me].name && strcmp(bt->dumps[me].name, why) == 0 )
       { print_trace(bt, me);
-	return;
+	done++;
       }
       if ( --me == bt->current-1 )
 	break;
     }
   }
 
-  Sdprintf("No backtrace named %s\n", why);
+  if ( !done )
+    Sdprintf("No backtrace named %s\n", why);
 }
 
 #endif /*HAVE_LIBUNWIND*/
@@ -440,7 +443,9 @@ print_backtrace(int last)		/* 1..SAVE_TRACES */
 
 void
 bstore_print_backtrace_named(btrace *bt, const char *why)
-{ if ( bt )
+{ int done = 0;
+
+  if ( bt )
   { int me = bt->current-1;
 
     for(;;)
@@ -448,14 +453,15 @@ bstore_print_backtrace_named(btrace *bt, const char *why)
 	me += SAVE_TRACES;
       if ( bt->why[me] && strcmp(bt->why[me], why) == 0 )
       { print_trace(bt, me);
-	return;
+	done++;
       }
       if ( --me == bt->current-1 )
 	break;
     }
   }
 
-  Sdprintf("No backtrace named %s\n", why);
+  if ( !done )
+    Sdprintf("No backtrace named %s\n", why);
 }
 
 
@@ -468,56 +474,23 @@ bstore_print_backtrace_named(btrace *bt, const char *why)
 
 #ifdef BTRACE_DONE
 
-#ifndef HAVE_CTIME_R
-#define ctime_r(timep, buf) strcpy(buf, ctime(timep))
-#endif
-
-static void
-crashHandler(int sig)
-{ int tid;
-  atom_t alias;
-  const pl_wchar_t *name = L"";
-  time_t now = time(NULL);
-  char tbuf[48];
-
-  signal(sig, SIG_DFL);
-  tid = PL_thread_self();
-  ctime_r(&now, tbuf);
-  tbuf[24] = '\0';
-
-  if ( PL_get_thread_alias(tid, &alias) )
-    name = PL_atom_wchars(alias, NULL);
-
-  Sdprintf("\nSWI-Prolog [thread %d (%Ws) at %s]: "
-	   "received fatal signal %d (%s)\n",
-	   PL_thread_self(), name, tbuf, sig, signal_name(sig));
-  print_c_backtrace("crash");
-  run_on_halt(&GD->os.exit_hooks, 4);
-
-#if defined(HAVE_KILL) && defined(HAVE_GETPID)
-  kill(getpid(), sig);
-#else
-  abort();
-#endif
-}
-
 void
 initBackTrace(void)
 {
 #ifdef SIGSEGV
-  PL_signal(SIGSEGV, crashHandler);
+  PL_signal(SIGSEGV, sigCrashHandler);
 #endif
 #ifdef SIGILL
-  PL_signal(SIGILL, crashHandler);
+  PL_signal(SIGILL, sigCrashHandler);
 #endif
 #if defined(SIGBUS) && SIGBUS != SIGSEGV
-  PL_signal(SIGBUS, crashHandler);
+  PL_signal(SIGBUS, sigCrashHandler);
 #endif
 #ifdef SIGFPE
-  PL_signal(SIGFPE, crashHandler);
+  PL_signal(SIGFPE, sigCrashHandler);
 #endif
 #ifdef SIGSYS
-  PL_signal(SIGSYS, crashHandler);
+  PL_signal(SIGSYS, sigCrashHandler);
 #endif
 }
 
@@ -853,7 +826,9 @@ print_backtrace(int last)		/* 1..SAVE_TRACES */
 
 void
 bstore_print_backtrace_named(btrace *bt, const char *why)
-{ if ( bt )
+{ int done = 0;
+
+  if ( bt )
   { int me = bt->current-1;
 
     for(;;)
@@ -861,14 +836,15 @@ bstore_print_backtrace_named(btrace *bt, const char *why)
 	me += SAVE_TRACES;
       if ( bt->dumps[me].name && strcmp(bt->dumps[me].name, why) == 0 )
       { print_trace(bt, me);
-	return;
+	done++;
       }
       if ( --me == bt->current-1 )
 	break;
     }
   }
 
-  Sdprintf("No backtrace named %s\n", why);
+  if ( !done )
+    Sdprintf("No backtrace named %s\n", why);
 }
 
 static LONG WINAPI crashHandler(PEXCEPTION_POINTERS pExceptionInfo)
@@ -891,6 +867,10 @@ initBackTrace(void)
 		 *	     SHARED		*
 		 *******************************/
 
+#ifndef HAVE_CTIME_R
+#define ctime_r(timep, buf) strcpy(buf, ctime(timep))
+#endif
+
 #ifdef BTRACE_DONE
 
 void
@@ -908,8 +888,95 @@ print_c_backtrace(const char *why)
     btrace_destroy(bt);
 }
 
+void
+sigCrashHandler(int sig)
+{ int tid;
+  atom_t alias;
+  const pl_wchar_t *name = L"";
+  time_t now = time(NULL);
+  char tbuf[48];
+
+  signal(sig, SIG_DFL);
+  tid = PL_thread_self();
+  ctime_r(&now, tbuf);
+  tbuf[24] = '\0';
+
+  if ( PL_get_thread_alias(tid, &alias) )
+    name = PL_atom_wchars(alias, NULL);
+
+  Sdprintf("\nSWI-Prolog [thread %d (%Ws) at %s]: "
+	   "received fatal signal %d (%s)\n",
+	   PL_thread_self(), name, tbuf, sig, signal_name(sig));
+  print_c_backtrace("crash");
+  Sdprintf("Running on_halt hooks with status %d\n", 128+sig);
+  run_on_halt(&GD->os.exit_hooks, 128+sig);
+
+#ifdef HAVE_KILL
+{ int pid;
+# ifdef O_PLMT
+#  ifdef HAVE_GETTID_SYSCALL
+#   ifdef HAVE_SYS_SYSCALL_H
+#    include <sys/syscall.h>
+#   endif
+  pid = syscall(__NR_gettid);
+#  elif defined(HAVE_GETTID_MACRO)
+  pid = gettid();
+#  else
+  pid = getpid();
+#  endif
+# else /*O_PLMT*/
+  pid = getpid();
+# endif
+  Sdprintf("Killing %d with default signal handlers\n", pid);
+  kill(pid, sig);
+}
+#else /*HAVE_KILL*/
+  Sdprintf("Aborting\n");
+  abort();
+#endif
+}
+
 #endif /*BTRACE_DONE*/
 
+		 /*******************************
+		 *	PROLOG CONNECTION	*
+		 *******************************/
+
+#if defined(O_DEBUG) && defined(BTRACE_DONE)
+#define BTRACE_PREDS 1
+
+static
+PRED_IMPL("c_backtrace_clear", 0, c_backtrace_clear, 0)
+{ PRED_LD
+
+  if ( LD->btrace_store )
+  { btrace_destroy(LD->btrace_store);
+    LD->btrace_store = NULL;
+  }
+
+  return TRUE;
+}
+
+static
+PRED_IMPL("c_backtrace_print", 1, c_backtrace_print, 0)
+{ char *s;
+
+  if ( PL_get_chars(A1, &s, CVT_ATOM|CVT_STRING|CVT_EXCEPTION) )
+  { print_backtrace_named(s);
+    return TRUE;
+  }
+
+  return FALSE;
+}
+
+#endif
+
+BeginPredDefs(cbtrace)
+#ifdef BTRACE_PREDS
+  PRED_DEF("c_backtrace_clear", 0, c_backtrace_clear, 0)
+  PRED_DEF("c_backtrace_print", 1, c_backtrace_print, 0)
+#endif
+EndPredDefs
 
 	         /*******************************
 		 *   FALLBACK IMPLEMENTATION	*
@@ -950,4 +1017,42 @@ initBackTrace(void)
 {
 }
 
+#ifdef HAVE_SIGNAL
+void
+sigCrashHandler(int sig)
+{ int tid;
+  atom_t alias;
+  const pl_wchar_t *name = L"";
+  time_t now = time(NULL);
+  char tbuf[48];
+
+  signal(sig, SIG_DFL);
+  tid = PL_thread_self();
+  ctime_r(&now, tbuf);
+  tbuf[24] = '\0';
+
+  if ( PL_get_thread_alias(tid, &alias) )
+    name = PL_atom_wchars(alias, NULL);
+
+  Sdprintf("\nSWI-Prolog [thread %d (%Ws) at %s]: "
+	   "received fatal signal %d (%s)\n",
+	   PL_thread_self(), name, tbuf, sig, signal_name(sig));
+  run_on_halt(&GD->os.exit_hooks, 4);
+
+#if defined(HAVE_KILL) && defined(HAVE_GETPID)
+  kill(getpid(), sig);
+#else
+  abort();
+#endif
+}
+
+#else
+
+void
+sigCrashHandler(int sig)
+{ fatalError("Something went wrong");
+}
+
+#endif /*HAVE_SIGNAL*/
+
 #endif /*BTRACE_DONE*/
