@@ -91,8 +91,12 @@ lookupProcedure(functor_t f, Module m)
   def->functor = valueFunctor(f);
   def->module  = m;
   def->shared  = 1;
-  def->args    = allocHeapOrHalt(sizeof(arg_info)*def->functor->arity);
-  memset(def->args, 0, sizeof(arg_info)*def->functor->arity);
+  if ( def->functor->arity > 0 )
+  { def->impl.any.args = allocHeapOrHalt(sizeof(arg_info)*def->functor->arity);
+    memset(def->impl.any.args, 0, sizeof(arg_info)*def->functor->arity);
+  } else
+  { def->impl.any.args = NULL;
+  }
   resetProcedure(proc, TRUE);
 
   DEBUG(MSG_PROC_COUNT, Sdprintf("Created %s\n", procedureName(proc)));
@@ -159,7 +163,7 @@ destroyDefinition(Definition def)
   freeCodesDefinition(def, FALSE);
 
   if ( false(def, P_FOREIGN|P_THREAD_LOCAL) )	/* normal Prolog predicate */
-  { freeHeap(def->args, sizeof(arg_info)*def->functor->arity);
+  { freeHeap(def->impl.any.args, sizeof(arg_info)*def->functor->arity);
     removeClausesPredicate(def, 0, FALSE);
     DEBUG(MSG_CGC_PRED,
 	  Sdprintf("destroyDefinition(%s)\n", predicateName(def)));
@@ -175,7 +179,9 @@ destroyDefinition(Definition def)
   { DEBUG(MSG_PROC_COUNT, Sdprintf("Unalloc foreign/thread-local: %s\n",
 				   predicateName(def)));
     if ( true(def, P_DIRTYREG) )
-    { DEBUG(0, Sdprintf("Dirty: %s\n", predicateName(def)));
+    { DEBUG(0, if ( GD->cleaning == CLN_NORMAL )
+	         Sdprintf("Destroying dirty predicate: %s\n",
+			  predicateName(def)));
       unregisterDirtyDefinition(def);
     }
 
@@ -263,7 +269,7 @@ resetProcedure(Procedure proc, bool isnew)
 { Definition def = proc->definition;
 
   if ( (true(def, P_DYNAMIC) /*&& def->references == 0*/) ||
-       !def->impl.any )
+       !def->impl.any.defined )
     isnew = TRUE;
 
   def->flags ^= def->flags & ~(SPY_ME|P_DIRTYREG);
@@ -272,7 +278,7 @@ resetProcedure(Procedure proc, bool isnew)
   def->impl.clauses.number_of_clauses = 0;
 
   if ( isnew )
-  { deleteIndexes(def, TRUE);
+  { deleteIndexes(&def->impl.clauses, TRUE);
     freeCodesDefinition(def, FALSE);
   } else
     freeCodesDefinition(def, TRUE);	/* carefully sets to S_VIRGIN */
@@ -428,6 +434,7 @@ void
 shareDefinition(Definition def)
 { int shared = ATOMIC_INC(&def->shared);
   assert(shared > 0);
+  (void)shared;
 }
 
 
@@ -1173,8 +1180,7 @@ The `where` argument is one of
   - CL_START (asserta)
   - CL_END   (assertz)
   - The clause reference before which the clause must be inserted.
-    This is used by reconsult.  Note that addClauseToIndexes() may
-    not add the clause to the index, but we do not care.
+    This is used by reconsult.
 
 (*) This function updates the indexing information.  If we have a static
 procedure, it deletes the supervisor. This is  probably a bit rough, but
@@ -1261,7 +1267,7 @@ abolishProcedure(Procedure proc, Module module)
 
     memset(ndef, 0, sizeof(*ndef));
     ndef->functor            = def->functor; /* should be merged with */
-    ndef->args		     = allocHeapOrHalt(sizeof(*ndef->args)*
+    ndef->impl.any.args	     = allocHeapOrHalt(sizeof(*ndef->impl.any.args)*
 					       def->functor->arity);
     ndef->module             = module;	     /* lookupProcedure()!! */
     ndef->codes		     = SUPERVISOR(virgin);
@@ -1374,6 +1380,8 @@ retractClauseDefinition(Definition def, Clause clause)
   deleteActiveClauseFromIndexes(def, clause); /* just updates "dirtyness" */
   def->impl.clauses.number_of_clauses--;
   def->impl.clauses.erased_clauses++;
+  if ( false(clause, UNIT_CLAUSE) )
+    def->impl.clauses.number_of_rules--;
 #ifdef O_LOGICAL_UPDATE
   clause->generation.erased = next_global_generation();
   setLastModifiedPredicate(def, clause->generation.erased);
@@ -1513,14 +1521,23 @@ find_prev(Definition def, ClauseRef prev, ClauseRef cref)
 
 
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+(*) This used to be acquire_def(def),  but announceErasedClause may call
+Prolog, leading to nested acquired definition. This is not needed anyway
+as the acquired definition is only  used   by  clause  GC, we are inside
+clause GC and clause GC calls cannot run in parallel.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 static size_t
 cleanDefinition(Definition def, DirtyDefInfo ddi, gen_t start, int *rcp)
-{ GET_LD
-  size_t removed = 0;
+{ size_t removed = 0;
   gen_t marked = ddi->oldest_generation;
   gen_t active = start < marked ? start : marked;
 
-  DEBUG(CHK_SECURE, checkDefinition(def));
+  DEBUG(CHK_SECURE,
+	LOCKDEF(def);
+	checkDefinition(def);
+        UNLOCKDEF(def));
 
   if ( mustCleanDefinition(def) )
   { ClauseRef cref, prev = NULL;
@@ -1528,7 +1545,7 @@ cleanDefinition(Definition def, DirtyDefInfo ddi, gen_t start, int *rcp)
     int left = 0;
 #endif
 
-    acquire_def(def);
+    assert(GD->clauses.cgc_active);		/* See (*) */
     for(cref = def->impl.clauses.first_clause;
 	cref && def->impl.clauses.erased_clauses;
 	cref=cref->next)
@@ -1561,13 +1578,15 @@ cleanDefinition(Definition def, DirtyDefInfo ddi, gen_t start, int *rcp)
     }
     if ( removed )
     { LOCKDEF(def);
-      cleanClauseIndexes(def, active);
+      cleanClauseIndexes(def, &def->impl.clauses, active);
       UNLOCKDEF(def);
     }
     freeLingeringDefinition(def, ddi);
-    release_def(def);
 
-    DEBUG(CHK_SECURE, checkDefinition(def));
+    DEBUG(CHK_SECURE,
+	  LOCKDEF(def);
+	  checkDefinition(def);
+	  UNLOCKDEF(def));
 
     DEBUG(MSG_PROC,
 	  Sdprintf("cleanDefinition(%s): removed %d, left %d, erased %d\n",
@@ -1688,7 +1707,7 @@ setMetapredicateMask(Definition def, arg_info *args)
 { size_t i, arity = def->functor->arity;
 
   for(i=0; i<arity; i++)
-    def->args[i].meta = args[i].meta;
+    def->impl.any.args[i].meta = args[i].meta;
 
   if ( isTransparentMetamask(def, args) )
     set(def, P_TRANSPARENT);
@@ -1781,7 +1800,7 @@ PRED_IMPL("meta_predicate", 1, meta_predicate, PL_FA_TRANSPARENT)
 static int
 unify_meta_argument(term_t head, Definition def, int i ARG_LD)
 { term_t arg = PL_new_term_ref();
-  int m = def->args[i].meta;
+  int m = def->impl.any.args[i].meta;
 
   _PL_get_arg(i+1, head, arg);
   if ( m < 10 )
@@ -1872,7 +1891,7 @@ PL_meta_predicate(predicate_t proc, const char *spec_s)
 	return FALSE;
     }
 
-    def->args[i].meta = spec;
+    def->impl.any.args[i].meta = spec;
     mask |= spec<<(i*4);
     if ( MA_NEEDS_TRANSPARENT(spec) )
       transparent = TRUE;
@@ -1893,7 +1912,7 @@ clear_meta_declaration(Definition def)
 { int i;
 
   for(i=0; i<def->functor->arity; i++)
-    def->args[i].meta = MA_ANY;
+    def->impl.any.args[i].meta = MA_ANY;
 
   clear(def, P_META|P_TRANSPARENT);
 }
@@ -2116,7 +2135,10 @@ pl_garbage_collect_clauses(void)
 		Definition def = n;
 #endif
 
-		DEBUG(CHK_SECURE, checkDefinition(def));
+		DEBUG(CHK_SECURE,
+		      LOCKDEF(def);
+		      checkDefinition(def);
+		      UNLOCKDEF(def));
 		ddi->oldest_generation = GEN_MAX; /* see (*) */
 	      });
 
@@ -2659,7 +2681,10 @@ pl_retractall(term_t head)
   setGenerationFrameVal(environment_frame, pushPredicateAccess(def));
   fid = PL_open_foreign_frame();
 
-  DEBUG(CHK_SECURE, checkDefinition(def));
+  DEBUG(CHK_SECURE,
+	LOCKDEF(def);
+	checkDefinition(def);
+        UNLOCKDEF(def));
   if ( allvars )
   { gen_t gen = generationFrame(environment_frame);
 
@@ -2702,6 +2727,11 @@ pl_retractall(term_t head)
   }
   popPredicateAccess(def);
   leaveDefinition(def);
+  DEBUG(CHK_SECURE,
+	LOCKDEF(def);
+	checkDefinition(def);
+	UNLOCKDEF(def));
+
   return endCritical;
 }
 
@@ -2819,13 +2849,14 @@ attribute_mask(atom_t key)
 }
 
 
-int
+static size_t
 num_visible_clauses(Definition def, atom_t key)
 { GET_LD;
 
   if ( LD->gen_reload != GEN_INVALID )
   { ClauseRef c;
-    int num_clauses = 0;
+    size_t num_clauses = 0;
+
     acquire_def(def);
     for(c = def->impl.clauses.first_clause; c; c = c->next)
     { Clause cl = c->value.clause;
@@ -2914,7 +2945,7 @@ pl_get_predicate_attribute(term_t pred,
   } else if ( key == ATOM_foreign )
   { return PL_unify_integer(value, true(def, P_FOREIGN) ? 1 : 0);
   } else if ( key == ATOM_number_of_clauses )
-  { int num_clauses;
+  { size_t num_clauses;
     if ( def->flags & P_FOREIGN )
       fail;
 
@@ -2922,10 +2953,12 @@ pl_get_predicate_attribute(term_t pred,
     num_clauses = num_visible_clauses(def, key);
     if ( num_clauses == 0 && false(def, P_DYNAMIC) )
       fail;
-
-    return PL_unify_integer(value, num_clauses);
+    return PL_unify_int64(value, num_clauses);
   } else if ( key == ATOM_last_modified_generation )
-  { return PL_unify_int64(value, def->last_modified);
+  { if ( def->flags & P_FOREIGN )
+      fail;
+    def = getProcDefinition(proc);
+    return PL_unify_int64(value, def->last_modified);
   } else if ( key == ATOM_number_of_rules )
   { if ( def->flags & P_FOREIGN )
       fail;
